##################################################################### 
# Copyright (C) 2007 Alex Clemesha <clemesha@gmail.com>
# 
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
##################################################################### 
#
# This code was inspired by the following sources:
# * The file 'guard.py' included in the Nevow Web Framework http://divmod.org/trac/wiki/DivmodNevow
# * The file 'guard.py' used in the Stiq website code http://www.stiq.it 

#twisted modules
from twisted.internet import task, defer
from twisted.cred.error import UnauthorizedLogin
from twisted.cred import credentials
from zope.interface import implements
from knoboo.external.twisted.web2 import iweb
from knoboo.external.twisted.web2 import server

#standard library
import random
import time
import md5


class Session(object):
    """A single users session, defined by the uid.
    """
    def __init__(self, uid, sessionManager):
        self.uid = uid #the cookie, a unique identifier (md5 hash)
        self.creationTime = self.lastAccessed = time.time()
        self.sessionManager = sessionManager
        self.authenticatedAs = None
        #self.setLifetime(60)
    
    def set_authCreds(self, creds):
        self.authenticatedAs = creds

    def get_authCreds(self):
        return self.authenticatedAs

    def get_uid(self):
        return self.uid

    def touch(self):
        self.lastAccessed = time.time()

    def expire(self):
        return defer.maybeDeferred(self.sessionManager.expiredSession, self)


class KnobooSession(Session):
    """A single users session that has associated running notebooks
    """

    def __init__(self, uid, sessionManager):
        Session.__init__(self, uid, sessionManager)
        self.uid = uid 
        self.creationTime = self.lastAccessed = time.time()
        self.sessionManager = sessionManager
        self.authenticatedAs = None
        self.kernels = set()

    def add_new_kernel(self, request):
        try:
            nbid = request.args['nbid'][0]
            self.kernels.add(nbid)
            return nbid
        except KeyError:
            return None

    def get_kernels(self):
        return self.kernels


class SessionsManager(object):
    """Manages all logged in users' sessions.
    """
    tickTime = 10000 #SESSION_CLEAN_FREQUENCY
    sessionLifetime = 1000000 #TRANSIENT_SESSION_LIFETIME
    sessionPersistentLifetime = 1000000 #PERSISTENT_SESSION_LIFETIME
    
    sessionFactory = KnobooSession
    
    def __init__(self):
        self.sessions = {}
        self.tick = task.LoopingCall(self._tick)

    def createSession(self):
        session = self.sessionFactory(self.createSessionID(), self)
        uid = session.get_uid()
        self.sessions[uid] = session
        if not self.tick.running and len(self.sessions) > 0:
            self.tick.start(self.tickTime)
        return session, uid
    
    def expiredSession(self, session):
        uid = session.get_uid()
        if uid in self.sessions:
            self.sessions.pop(uid)
        if self.tick.running and len(self.sessions) == 0:
            self.tick.stop()

    def getSession(self, uid):
        session = self.sessions.get(uid)
        if session:
            session.touch()
        return session

    def createSessionID(self):
        """Generate a new session ID."""
        data = "%s_%s" % (str(random.random()) , str(time.time()))
        return md5.new(data).hexdigest()

    def _tick(self):
        """Remove expired sessions.
        """
        now = time.time()
        for uid, session in self.sessions.items():
            age = now - session.lastAccessed
            max = self.sessionLifetime
            #if session.persistent:
            #    max = self.sessionPersistentLifetime
            if age > max:
                self.sessions[uid].expire()
        if not self.sessions and self.tick.running:
            self.tick.stop()

class MindManager(object):
    """Pass info from stateful to stateless"""
    def __init__(self, uid, segments, nbid=None, kill_kernels=[]): 
        self.cookie = uid
        self.segments = segments
        self.nbid = nbid
        self.kill_kernels = kill_kernels

    def is_kernel_operation(self):
        segs = self.segments
        if ('kernel' in segs) or ('completer' in segs) or ('eval' in segs):
            return True
        return False

class SessionWrapper(object):
    """Wrapper on top of a request to implement HTTP Sessions.

    This class implements iweb.IResource, which means it
    behaves exactly like a resource.Resource subclass, but
    it really only does session related logic, and then returns
    real resource depending on a users credentials.
    """
    implements(iweb.IResource)
    
    cookieManager = None
    mindFactory = MindManager
    sessionManager = SessionsManager()

    # The interface to cred for when logging into the portal
    credInterface = iweb.IResource

    def __init__(self, portal):
        self.portal = portal

    def renderHTTP(self, request):
        """This method should never be called.

        The reason for this is that SessionWrapper
        is only a 'wrapper' on top of real Resources 
        that actually handle rendering and interaction.
        """
        return http.Response(responsecode.NOT_FOUND)
       
    def locateChild(self, request, segments):
        """Serve Resources depending on users Authentication status.
        
        Decide the authentication status of a given user
        by first checking is the user is trying to log in,
        in which case get the username/password.  Else get
        the resource associated with the session cookie.
        """
        if segments and segments[0] == "login":
            #get the username and password in the postdata
            #the callback function needs no args because the parsing
            #of the POSTData just updates the request args.
            l = server.parsePOSTData(request)
            l.addCallback(lambda _: self.requestPasswordAthentication(request, segments))
            return l
        session = self.getSession(request)
        if session is None:
            return self.requestAnonymousAuthentication(request, segments)
        else:
            if segments and segments[0] == "logout":
                return self.logout(session, request, segments)
            else:
                if request.method == "POST":
                    d = server.parsePOSTData(request)
                    d.addCallback(lambda _: self.locateResource(request, segments, session))
                    return d
                return self.locateResource(request, segments, session)

    def locateResource(self, request, segments, session):
        """Locate the resource for an authenticated session.

        This method is used to actually get the resource that
        was requested after the request is passed through locateChild
        by sending the user's credentials to 'portal.login'.

        The user's session is checked in locateChild so at this
        point the user has already succesfully authenticated.
        """
        def _success(avatar, request, segments):
            iface, rsrc, logout = avatar
            return rsrc, segments
        creds = session.get_authCreds()
        cookie = session.get_uid()
        nbid = session.add_new_kernel(request)
        mind = self.mindFactory(cookie, segments, nbid=nbid)
        d = self.portal.login(creds, mind, self.credInterface)
        d.addCallback(_success, request, segments)
        return d

    def requestPasswordAthentication(self, request, segments):
        """
        Try to authenticate with a username and password from
        a web login form.  Depending on the given credentials,
        return a custom 'view' of protected resources.
        """
        creds = self.getCredentials(request)
        session, newCookie = self.sessionManager.createSession()
        mind = self.mindFactory(newCookie, segments)
        d = self.portal.login(creds, mind, self.credInterface)
        d.addCallback(self._loginSuccess, session, creds, segments)
        d.addErrback(self._loginFailure, request, segments)
        return d 

    def requestAnonymousAuthentication(self, request, segments):
        """
        Anonymous authentication, the user can only
        access non-protected resources.
        """
        def _success(avatar, request, segments):
            iface, resource, logout = avatar
            return resource, segments
        creds = credentials.Anonymous() #anonymous user.
        #session, newCookie = self.sessionManager.createSession()
        mind = None 
        d = self.portal.login(creds, mind, self.credInterface)
        d.addCallback(_success, request, segments)
        return d

    def getSession(self, request):
        """Get a user's session defined by the request cookie.

        Pull the cookie out of the http header and
        pass it to the session manager, which handles
        finding the user's actual session object.
        """
        cookiedata = request.headers.getHeader('cookie')
        cookie = ''
        if cookiedata:
            for c in cookiedata:
                if c.name == 'sid':
                    cookie = c.value
        session = self.sessionManager.getSession(cookie)
        return session
 
    def getCredentials(self, request):
        "Get username, password from post args."
        username = request.args.get('email', [''])[0]
        password = request.args.get('password', [''])[0]
        return credentials.UsernamePassword(username, password)

    def _loginSuccess(self, (iface, rsrc, logout), session, creds, segments):
        """Return resource after login success.
        
        Also saved the user's credentials to later associate 
        them with the users session.
        """
        session.set_authCreds(creds)
        return rsrc, () #immediately render
    
    def _loginFailure(self, avatar, request, segments):
        """Login failed.
        Return the resource that will handle the fact that
        the login just failed.
        """
        return self.requestAnonymousAuthentication(request, segments)
                 
    def logout(self, session, request, segments):
        """Destroy the users session.

        This pops the uid that defines the users session
        out of the sessions dict.  A new anonymous session
        is immediatly created for this user.
        """
        def _success(avatar, request, segments):
            iface, resource, logout = avatar
            self.sessionManager.expiredSession(session)
            return self.requestAnonymousAuthentication(request, segments)
        cookie = session.get_uid()
        k = session.get_kernels()
        mind = self.mindFactory(cookie, segments, kill_kernels=k)
        creds = session.get_authCreds()
        d = self.portal.login(creds, mind, self.credInterface)
        d.addCallback(_success, request, segments)
        return d



