##################################################################### 
# Copyright (C) 2007 Dorian Raymer <deldotdr@gmail.com>
# 
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
##################################################################### 

from twisted.internet import protocol
from twisted.internet import reactor, defer
from twisted.internet import error
#from twisted.web import xmlrpc
from twisted.python import log
import socket
import time
import os
import sys

# local import
import xmlrpc 
import skeleton_python_kernel as kernelMaker

# Configuration
from knoboo import config
kbconfig = config.Config()
cfg = kbconfig.read_config()
RUNPATH = str(cfg['kernel_env'])

# Global parameters:
# Port range to use for Kernel Worker Server to use
HOST = "localhost"
MIN_PORT = 9000
MAX_PORT = 10000


# Kernel interrupt parameters
INT_INTERVAL = 0.01
INT_MAX_TRYS = 20


def getUnboundPort(host, min_port=MIN_PORT, max_port=MAX_PORT):
    """Find an unbound port by attempting to bind to one with in the
    given range. On success, unbind and return the port number.
    TODO: This is blocking. This could be redone in a 'Twisted' async way (look at
    example in begining of oreilly book)
    maybe defer this to a thread...
    """
    sock = socket.socket()
    for port in range(min_port, max_port):
        try:
            sock.bind((host, port))
            sock.close()
            return port
        except socket.error:
            pass
    # in the unlikely situation that no ports were available, return 0?
    return 0


class KernelProcessProtocol(protocol.ProcessProtocol):
    """
    Subclass for controlling a computational kernel process.
    """
    def __init__(self):
        #log.msg("KernelProcessProtocol Object Instantiated")
        self.running = 0

    def connectionMade(self):
        """A good connection indicates the kernel is running.
        """
        self.running = 1
        log.msg("kernel started.")

    def outReceived(self, data):
        #log.msg("kernel outReceived with %d bytes"%len(data))
        log.msg("Kernel out %s"%data)
        #pass

    def errReceived(self, data):
        #log.msg("kernel errReceived with %d bytes"%len(data))
        log.msg("Kernel error %s"%data)
        #pass
 
    def inConnectionLost(self):
        #m = "kernel in connection lost"
        #log.msg(m)
        pass

    def outConnectionLost(self):
        #m = "kernel out connection lost. The child closed their stdout"
        #log.msg(m)
        pass
 
    def errConnectionLost(self):
        #m = "kernel err connection lost. the child closef their stderr"
        #log.msg(m)
        pass

    def processEnded(self, status_object):
        """processEnded indicates the kernel has stopped running.
        """
        self.running = 0
        try:
            m = "kernel process ended, status %d"%status_object.value.exitCode
            log.msg(m)
        except TypeError:
            m = "kernel process ended. non-int exitCode"
            log.msg(m)
        #log.msg("quitting Kernel, Bye!")



class KernelManager(object):
    """Creates, controls, and communicates with a remote kernel process.
    Creation: a kernel file is generated and written to a specific location
            in the file system.

    Control: The process is started through the processProtocol extension
            KernelProcessProtocol. The processProtocol provides a transport
            object for signaling the process. 

    Communication: The computational functionality is interfaced through
            XML-RPC. Standard basic methods are called by name and any
            arguments are passed as strings. 
            
    """

    def __init__(self, system_name, nbid, host=HOST):
        self.system_name = system_name
        self.kernel_file_name = str(nbid)
        self.host = host
        self.port = getUnboundPort(host)
        self.runpath = RUNPATH
        self._write_kernel_server_file()
        self.setup_process()
        self.start()
        self.kernelurl = "http://" + self.host + ":" + str(self.port) + "/"
        #self.kernelConnection = xmlrpc.KernelProxy(self.kernelurl)
        self.kernelConnection = xmlrpc.KernelClient(self.kernelurl)


    def _write_kernel_server_file(self):
        """This is a zero-order start-of-concept implementation of creating
        and running the computation process in a special *SECURE/SAFE*
        environment like a chroot jail or virtual file system...
        """
        kernelMaker.generate_kernel_server_file(self.runpath, self.kernel_file_name, self.port)

    def setup_process(self):
        runpath = self.runpath + self.kernel_file_name 
        pythonpath = os.getenv('PYTHONPATH')
        #pidfile = self.kernel_file_name + '.pid'
        #executable = sys.argv[0] #the twistd script
        #args = ["twistd", "-ny", self.kernel_file_name, "-l", "-", "--pidfile", pidfile]
        self.kernelProcess = KernelProcessProtocol() # maybe kernelProtocol?
        self.executable = "python"
        self.args = ("python", self.kernel_file_name)       
        self.env = {'PYTHONPATH':pythonpath, 'PATH':os.getenv('PATH')}
        #self.env = {'PATH':os.getenv('PATH')}
        self.path = self.runpath + self.kernel_file_name #change kernel_file_name 
        self.uid = None
        self.gid = None
        self.usePTY = 0
        self.childFDs = None

    def start(self):
        """Start the kernel by spawning a kernel process.
        The arguments of spawnProcess are specific to instances of this
        classs and are set at instantiation by self.setup_process.
        """
        self.computing = 0
        try:
            self.kernelspawned = reactor.spawnProcess(self.kernelProcess,
                                                self.executable,
                                                args=self.args,
                                                env=self.env,
                                                path=self.path,
                                                uid=self.uid,
                                                gid=self.gid,
                                                usePTY=self.usePTY,
                                                childFDs=self.childFDs)
        except OSError, e:
            print 'SPAWN ERROR',e
            raise

    def quit(self):
        """Cleanly kill a kernel process.
        """
        pass

    def interrupt(self):
        """Send kernel process SIGINT if a computation is running.
        (If a computation is not running, the signal will get caught by the
        rpc server and it will shut down!)
        """
        trys = 0
        self.dinterupt = defer.Deferred()
        self.intcalllater = reactor.callLater(INT_INTERVAL, self._check_interrupt, trys)
        self.dinterupt.addCallback(self._cancel_interrupt)
        self.dinterupt.addErrback(self._interruptErrback)
        return self.dinterupt

    def kill(self):
        """Send kernel process SIGKILL. 
        
        return True, indicating the kernel is killed
        TODO: this could be improved when the kernel manager API is
        more firmly established.
        Also, might want to implement a way to confirm the process was
        killed.
        """
        self.kernelProcess.transport.signalProcess("KILL")
        return True

    def pid(self):
        return self.kernelspawned.pid

    def eval(self, input):
        if not self.kernelProcess.running:
            self.start()
        conn = self.kernelConnection.callRemote("evalinput", input)
        self.computing += 1
        conn.addCallback(self._evalCallback)
        conn.addErrback(self._evalErrback)
        #TODO Need to add an errback for lost connection (which would
        # happen in the event of a killed kernel)
        return conn

    def complete_name(self, input):
        if not self.kernelProcess.running:
            self.start()
        return self.kernelConnection.callRemote("completename", input)
    
    def complete_attr(self, input):
        if not self.kernelProcess.running:
            self.start()
        return self.kernelConnection.callRemote("completeattr", input)


    def _evalCallback(self, result):
        self.computing -= 1
#        try:
#            if self.intcalllater.active():
#                self.intcalllater.cancel()
#                self.dinterupt.callback('eval reponse received')
#        except error.AlreadyCalled:
#            print 'caught AlreacyCalled #####'
#        except error.AlreadyCancelled:
#            print 'caught AlreadyCancelled $$$$$$$$$'
#        except AttributeError:
#            print 'caught Attribute Error %5%'
        # finally:
        return result

    def _evalErrback(self, res):
        """Triggered by query factory if connection is lost.
        Probably means kernel was killed.
        Need to know all possible failure types and sources!
        """
        print 'evalErrback @@@',res
        #err =res.trap(error.ConnectionDone)
        #if err == error.ConnectionDone:
        #    return 'Kernel connection Lost'
        return res


    def _check_interrupt(self, trys):
        trys += 1
        if self.computing and trys < INT_MAX_TRYS:
            self.kernelProcess.transport.signalProcess("INT")
            self.intcalllater = reactor.callLater(INT_INTERVAL, self._check_interrupt, trys)
        elif trys >= INT_MAX_TRYS:
            self.dinterupt.errback(False)
        else:
            self.dinterupt.callback('computation stopped: %s trys'%str(trys))

    def _cancel_interrupt(self, status):
        """Send the kernel a request to resume normal mode.
        This request will follow any eval requests made after the one being
        interrupted.
        Upon interrpt, an interrupt flag is set in the kernel to prevent
        any pending eval requests from actually being computed.
        """
        conn = self.kernelConnection.callRemote("cancelinterrupt")
        conn.addCallback(self._cancel_interruptCallback)
        return conn

    def _cancel_interruptCallback(self, res):
        return res

    def _interruptErrback(self, status):
        print 'InterruptErrorback!!!', status
        return status











