"""
XML-RPC proxy support.

This is completely based off of t.w.xmlrpc.

The difference is the queryFactory inherits t.i.p.ReconnectingClientFactory
instead of just ClientFactory.

This is designed for communicating with a kernel xml-rpc server that
completely blocks while a request is processed.

The main benefit of using the reconnecting ClientFactory is during kernel
startup; the kernel control/manager has no way of knowing when the kernel
server is ready to receive requests.
"""

import xmlrpclib
import urlparse

from twisted.internet import defer, protocol, reactor
from knoboo.external.twisted.web2.xmlrpc import QueryProtocol, payloadTemplate
from knoboo.external.twisted.web2.xmlrpc import Client
from twisted.python import failure

class KernelQueryFactory(protocol.ReconnectingClientFactory):

    deferred = None
    protocol = QueryProtocol
    maxRetries = 20

    def __init__(self, path, host, method, user=None, password=None, *args):
        self.path, self.host = path, host
        self.user, self.password = user, password
        self.payload = payloadTemplate % (method, xmlrpclib.dumps(args))
        self.deferred = defer.Deferred()

    def parseResponse(self, contents):
        if not self.deferred:
            return
        self.resetDelay()
        try:
            response = xmlrpclib.loads(contents)
        except:
            print 'parseRESPONSE Error ############'
            self.deferred.errback(failure.Failure())
            self.deferred = None
        else:
            self.deferred.callback(response[0][0])
            self.deferred = None

    def clientConnectionLost(self, _, reason):
        """ReconnectingClientFactory has this method try to reconnect
        automatically. So far, we only want this to happen on the initial
        connection.
        """
        if self.deferred is not None:
            print 'CONECTION LOST', reason
            self.deferred.errback(reason)
            self.deferred = None

    def badStatus(self, status, message):
        """need to fix errors differing between web and web2
        """
        #self.deferred.errback(ValueError(status, message))
        self.deferred.errback(BadResponseCodeError(status, message))
        self.deferred = None

class KernelClient(Client):
    def _buildFactory(self, method, *args):
        return KernelQueryFactory(self.path, self.host, method, self.user,
                self.password, *args)



class XKernelClient(object):
    """A Client for making remote XML-RPC calls.

    Pass the URL of the remote XML-RPC server to the constructor.

    Use proxy.callRemote('foobar', *args) to call remote method
    'foobar' with *args.

    """

    def __init__(self, url, user=None, password=None):
        """
        @type url: C{str}
        @param url: The URL to which to post method calls.  Calls will be made
        over SSL if the scheme is HTTPS.  If netloc contains username or
        password information, these will be used to authenticate, as long as
        the C{user} and C{password} arguments are not specified.

        @type user: C{str} or None
        @param user: The username with which to authenticate with the server
        when making calls.  If specified, overrides any username information
        embedded in C{url}.  If not specified, a value may be taken from C{url}
        if present.

        @type password: C{str} or None
        @param password: The password with which to authenticate with the
        server when making calls.  If specified, overrides any password
        information embedded in C{url}.  If not specified, a value may be taken
        from C{url} if present.
        """
        scheme, netloc, path, params, query, fragment = urlparse.urlparse(url)
        netlocParts = netloc.split('@')
        if len(netlocParts) == 2:
            userpass = netlocParts.pop(0).split(':')
            self.user = userpass.pop(0)
            try:
                self.password = userpass.pop(0)
            except:
                self.password = None
        else:
            self.user = self.password = None
        hostport = netlocParts[0].split(':')
        self.host = hostport.pop(0)
        try:
            self.port = int(hostport.pop(0))
        except:
            self.port = None
        self.path = path
        if self.path in ['', None]:
            self.path = '/'
        self.secure = (scheme == 'https')
        if user is not None:
            self.user = user
        if password is not None:
            self.password = password
    
    def _buildFactory(self, method, *args):
        return QueryFactory(self.path, self.host, method, self.user,
                    self.password, *args)

    def callRemote(self, method, *args):
        factory = self._buildFactory(method, *args)
        if self.secure:
            from twisted.internet import ssl
            reactor.connectSSL(self.host, self.port or 443,
                               factory, ssl.ClientContextFactory())
        else:
            reactor.connectTCP(self.host, self.port or 80, factory)
        return factory.deferred


#
#class _QueryFactory(protocol.ReconnectingClientFactory):
#    deferred = None
#    protocol = QueryProtocol
#    maxRetries = 20
#
#    def __init__(self, path, host, method, user=None, password=None,
#                 allowNone=False, args=()):
#        self.path, self.host = path, host
#        self.user, self.password = user, password
#        self.payload = payloadTemplate % (method,
#            xmlrpclib.dumps(args, allow_none=allowNone))
#        self.deferred = defer.Deferred()
#
#    def parseResponse(self, contents):
#        if not self.deferred:
#            return
#        self.resetDelay()
#        try:
#            response = xmlrpclib.loads(contents)
#        except:
#            # Could pass an exception...Maybe a custom knoboo one
#            print 'parseRESPONSE Error ############'
#            self.deferred.errback(failure.Failure())
#            self.deferred = None
#        else:
#            self.deferred.callback(response[0][0])
#            self.deferred = None
#        # finally:
#        #     self.resetDelay()
#
#    def clientConnectionLost(self, _, reason):
#        if self.deferred is not None:
#            print 'CONECTION LOST', reason
#            self.deferred.errback(reason)
#            self.deferred = None
#
#
#    def badStatus(self, status, message):
#        print '~~~~BAD STATUS',status,message
#        self.deferred.errback(ValueError(status, message))
#        self.deferred = None
#
#
#
#class KernelProxy:
#    """
#    Specific implementation of a Proxy for making remote XML-RPC calls to a
#    computation kernel server.
#    """
#
#    queryFactory = _QueryFactory
#
#    def __init__(self, url, user=None, password=None, allowNone=False):
#        """
#        @type url: C{str}
#        @param url: The URL to which to post method calls.  Calls will be made
#        over SSL if the scheme is HTTPS.  If netloc contains username or
#        password information, these will be used to authenticate, as long as
#        the C{user} and C{password} arguments are not specified.
#
#        @type user: C{str} or None
#        @param user: The username with which to authenticate with the server
#        when making calls.  If specified, overrides any username information
#        embedded in C{url}.  If not specified, a value may be taken from C{url}
#        if present.
#
#        @type password: C{str} or None
#        @param password: The password with which to authenticate with the
#        server when making calls.  If specified, overrides any password
#        information embedded in C{url}.  If not specified, a value may be taken
#        from C{url} if present.
#
#        @type allowNone: C{bool} or None
#        @param allowNone: allow the use of None values in parameters. It's
#        passed to the underlying xmlrpclib implementation. Default to False.
#        """
#        scheme, netloc, path, params, query, fragment = urlparse.urlparse(url)
#        netlocParts = netloc.split('@')
#        if len(netlocParts) == 2:
#            userpass = netlocParts.pop(0).split(':')
#            self.user = userpass.pop(0)
#            try:
#                self.password = userpass.pop(0)
#            except:
#                self.password = None
#        else:
#            self.user = self.password = None
#        hostport = netlocParts[0].split(':')
#        self.host = hostport.pop(0)
#        try:
#            self.port = int(hostport.pop(0))
#        except:
#            self.port = None
#        self.path = path
#        if self.path in ['', None]:
#            self.path = '/'
#        self.secure = (scheme == 'https')
#        if user is not None:
#            self.user = user
#        if password is not None:
#            self.password = password
#        self.allowNone = allowNone
#
#    def callRemote(self, method, *args):
#        factory = self.queryFactory(
#            self.path, self.host, method, self.user,
#            self.password, self.allowNone, args)
#        if self.secure:
#            from twisted.internet import ssl
#            reactor.connectSSL(self.host, self.port or 443,
#                               factory, ssl.ClientContextFactory())
#        else:
#            reactor.connectTCP(self.host, self.port or 80, factory)
#        return factory.deferred
#
